import sys
if sys.version_info[0] < 3:
    import inspect

    def count_args(fcn):
        return len(inspect.getargspec(fcn).args)
else:
    from inspect import signature

    def count_args(fcn):
        return len(signature(fcn).parameters)

import types

import torchbearer
from torchbearer.callbacks import Callback


class LambdaCallback(Callback):
    def __init__(self, func):
        self.func = func

    def on_lambda(self, state):
        return self.func(state)


def bind_to(target):
    def decorator(func):
        if isinstance(func, LambdaCallback):
            callback = func
        else:
            callback = LambdaCallback(func)
        setattr(callback, target.__name__, types.MethodType(lambda self, state: self.on_lambda(state), callback))
        return callback
    return decorator


def on_start(func):
    """ The :func:`on_start` decorator is used to initialise a :class:`.Callback` with :meth:`~.Callback.on_start`
    calling the decorated function

    Args:
        func (function): The function(state) to *decorate*

    Returns:
        Callback: Initialised callback with :meth:`~.Callback.on_start` calling func
    """
    return bind_to(Callback.on_start)(func)


def on_start_epoch(func):
    """ The :func:`on_start_epoch` decorator is used to initialise a :class:`.Callback` with
    :meth:`~.Callback.on_start_epoch` calling the decorated function

        Args:
        func (function): The function(state) to *decorate*

    Returns:
        Callback: Initialised callback with :meth:`~.Callback.on_start_epoch` calling func
    """
    return bind_to(Callback.on_start_epoch)(func)


def on_start_training(func):
    """ The :func:`on_start_training` decorator is used to initialise a :class:`.Callback` with :meth:`~.Callback.on_start_training`
    calling the decorated function

    Args:
        func (function): The function(state) to *decorate*

    Returns:
        Callback: Initialised callback with :meth:`~.Callback.on_start_training` calling func
    """
    return bind_to(Callback.on_start_training)(func)


def on_sample(func):
    """ The :func:`on_sample` decorator is used to initialise a :class:`.Callback` with :meth:`~.Callback.on_sample`
    calling the decorated function

    Args:
        func (function): The function(state) to *decorate*

    Returns:
        Callback: Initialised callback with :meth:`~.Callback.on_sample` calling func
    """
    return bind_to(Callback.on_sample)(func)


def on_forward(func):
    """ The :func:`on_forward` decorator is used to initialise a :class:`.Callback` with :meth:`~.Callback.on_forward`
    calling the decorated function

    Args:
        func (function): The function(state) to *decorate*

    Returns:
        Callback: Initialised callback with :meth:`~.Callback.on_forward` calling func
    """
    return bind_to(Callback.on_forward)(func)


def on_criterion(func):
    """ The :func:`on_criterion` decorator is used to initialise a :class:`.Callback` with :meth:`~.Callback.on_criterion`
    calling the decorated function

    Args:
        func (function): The function(state) to *decorate*

    Returns:
        Callback: Initialised callback with :meth:`~.Callback.on_criterion` calling func
    """
    return bind_to(Callback.on_criterion)(func)


def on_backward(func):
    """ The :func:`on_backward` decorator is used to initialise a :class:`.Callback` with :meth:`~.Callback.on_backward`
    calling the decorated function

    Args:
        func (function): The function(state) to *decorate*

    Returns:
        Callback: Initialised callback with :meth:`~.Callback.on_backward` calling func
    """
    return bind_to(Callback.on_backward)(func)


def on_step_training(func):
    """ The :func:`on_step_training` decorator is used to initialise a :class:`.Callback` with :meth:`~.Callback.on_step_training`
    calling the decorated function

    Args:
        func (function): The function(state) to *decorate*

    Returns:
        Callback: Initialised callback with :meth:`~.Callback.on_step_training` calling func
    """
    return bind_to(Callback.on_step_training)(func)


def on_end_training(func):
    """ The :func:`on_end_training` decorator is used to initialise a :class:`.Callback` with :meth:`~.Callback.on_end_training`
    calling the decorated function

    Args:
        func (function): The function(state) to *decorate*

    Returns:
        Callback: Initialised callback with :meth:`~.Callback.on_end_training` calling func
    """
    return bind_to(Callback.on_end_training)(func)


def on_end_epoch(func):
    """ The :func:`on_end_epoch` decorator is used to initialise a :class:`.Callback` with :meth:`~.Callback.on_end_epoch`
    calling the decorated function

    Args:
        func (function): The function(state) to *decorate*

    Returns:
        Callback: Initialised callback with :meth:`~.Callback.on_end_epoch` calling func
    """
    return bind_to(Callback.on_end_epoch)(func)


def on_end(func):
    """ The :func:`on_end` decorator is used to initialise a :class:`.Callback` with :meth:`~.Callback.on_end`
    calling the decorated function

    Args:
        func (function): The function(state) to *decorate*

    Returns:
        Callback: Initialised callback with :meth:`~.Callback.on_end` calling func
    """
    return bind_to(Callback.on_end)(func)


def on_start_validation(func):
    """ The :func:`on_start_validation` decorator is used to initialise a :class:`.Callback` with :meth:`~.Callback.on_start_validation`
    calling the decorated function

    Args:
        func (function): The function(state) to *decorate*

    Returns:
        Callback: Initialised callback with :meth:`~.Callback.on_start_validation` calling func
    """
    return bind_to(Callback.on_start_validation)(func)


def on_sample_validation(func):
    """ The :func:`on_sample_validation` decorator is used to initialise a :class:`.Callback` with :meth:`~.Callback.on_sample_validation`
    calling the decorated function

    Args:
        func (function): The function(state) to *decorate*

    Returns:
        Callback: Initialised callback with :meth:`~.Callback.on_sample_validation` calling func
    """
    return bind_to(Callback.on_sample_validation)(func)


def on_forward_validation(func):
    """ The :func:`on_forward_validation` decorator is used to initialise a :class:`.Callback` with :meth:`~.Callback.on_forward_validation`
    calling the decorated function

    Args:
        func (function): The function(state) to *decorate*

    Returns:
        Callback: Initialised callback with :meth:`~.Callback.on_forward_validation` calling func
    """
    return bind_to(Callback.on_forward_validation)(func)


def on_criterion_validation(func):
    """ The :func:`on_criterion_validation` decorator is used to initialise a :class:`.Callback` with :meth:`~.Callback.on_criterion_validation`
    calling the decorated function

    Args:
        func (function): The function(state) to *decorate*

    Returns:
        Callback: Initialised callback with :meth:`~.Callback.on_criterion_validation` calling func
    """
    return bind_to(Callback.on_criterion_validation)(func)


def on_end_validation(func):
    """ The :func:`on_end_validation` decorator is used to initialise a :class:`.Callback` with :meth:`~.Callback.on_end_validation`
    calling the decorated function

    Args:
        func (function): The function(state) to *decorate*

    Returns:
        Callback: Initialised callback with :meth:`~.Callback.on_end_validation` calling func
    """
    return bind_to(Callback.on_end_validation)(func)


def on_step_validation(func):
    """ The :func:`on_step_validation` decorator is used to initialise a :class:`.Callback` with :meth:`~.Callback.on_step_validation`
    calling the decorated function

    Args:
        func (function): The function(state) to *decorate*

    Returns:
        Callback: Initialised callback with :meth:`~.Callback.on_step_validation` calling func
    """
    return bind_to(Callback.on_step_validation)(func)


def on_checkpoint(func):
    """ The :func:`on_checkpoint` decorator is used to initialise a :class:`.Callback` with :meth:`~.Callback.on_checkpoint`
    calling the decorated function

    Args:
        func (function): The function(state) to *decorate*

    Returns:
        Callback: Initialised callback with :meth:`~.Callback.on_checkpoint` calling func
    """
    return bind_to(Callback.on_checkpoint)(func)


def add_to_loss(func):
    """ The :func:`add_to_loss` decorator is used to initialise a :class:`.Callback` with the value returned from func
    being added to the loss

    Args:
        func (function): The function(state) to *decorate*

    Returns:
        Callback: Initialised callback which adds the returned value from func to the loss
    """
    @on_criterion
    @on_criterion_validation
    def add_to_loss_callback(state):
        state[torchbearer.LOSS] = state[torchbearer.LOSS] + func(state)

    return add_to_loss_callback


def once(fcn):
    """
    Decorator to fire a callback once in the lifetime of the callback. If the callback is a class method, each
    instance of the class will fire only once. For functions, only the first instance will fire (even if more than
    one function is present in the callback list).

    Args:
        fcn (function): the `torchbearer callback` function to decorate.

    Returns:
        the decorator
    """
    def _once(self, _):
        try:
            return not self.__done__
        except AttributeError:
            self.__done__ = True
            return True

    return only_if(_once)(fcn)


def once_per_epoch(fcn):
    """Decorator to fire a callback once (on the first call) in any given epoch. If the callback is a class method, each
    instance of the class will fire once per epoch. For functions, only the first instance will fire (even if more than
    one function is present in the callback list).

    .. note::
        The decorated callback may exhibit unusual behaviour if it is reused

    Args:
        fcn (function): the `torchbearer callback` function to decorate.

    Returns:
        the decorator
    """
    def ope(self, state):
        try:
            if state[torchbearer.EPOCH] != self.__last_epoch__:
                self.__last_epoch__ = state[torchbearer.EPOCH]
                return True
            return False
        except AttributeError:
            self.__last_epoch__ = state[torchbearer.EPOCH]
            return True

    return only_if(ope)(fcn)


def only_if(condition_expr):
    """
    Decorator to fire a callback only if the given conditional expression function returns True. The conditional
    expression can be a function of state or self and state. If the decorated function is not a class method (i.e. it
    does not take state) the decorated function will be passed instead. This enables the storing of temporary variables.

    Args:
        condition_expr (function(self, state) or function(self)): a function/lambda which takes state and optionally\
                        self that must evaluate to true for the decorated `torchbearer callback` to be called. The\
                        `state` object passed to the callback will be passed as an argument to the condition function.

    Returns:
        the decorator
    """
    def condition_decorator(fcn):
        if isinstance(fcn, LambdaCallback):
            fcn.func = condition_decorator(fcn.func)
            return fcn
        else:
            count = count_args(fcn)
            if count == 2 and not hasattr(fcn, '__self__'):  # Assume Class method
                def decfcn(o, state):
                    try:
                        res = condition_expr(o, state)
                    except TypeError:
                        res = condition_expr(state)
                    if res:
                        return fcn(o, state)
            else:  # Assume function of state
                def id_fcn(state):
                    return fcn(state)  # Hack to allow setting attributes of bound methods

                def decfcn(state):
                    try:
                        res = condition_expr(id_fcn, state)
                    except TypeError:
                        res = condition_expr(state)
                    if res:
                        return id_fcn(state)
            return decfcn
    return condition_decorator



import torchbearer as tb
from torchbearer import Callback
import numpy as np
from torchbearer import cite

bibtex = """
@inproceedings{smith2017cyclical,
  title={Cyclical learning rates for training neural networks},
  author={Smith, Leslie N},
  booktitle={2017 IEEE Winter Conference on Applications of Computer Vision (WACV)},
  pages={464--472},
  year={2017},
  organization={IEEE}
}
"""


@cite(bibtex)
class CyclicLR(Callback):
    """ Learning rate finder that cyclicly varies the rate. Based off of the keras implementation referenced in the `paper <https://arxiv.org/abs/1506.01186>`_.

    Args:
        base_lr (float / list): Float or list of floats for the base (min) learning rate for each optimiser parameter group
        max_lr (float / list):Float or list of floats for the max learning rate for each optimiser parameter group
        step_size (int / list): int or list of ints for the step size (half cyclic period) for each optimiser parameter group
        mode (str): One of (triangular, triangular2, exp_range) - the mode to use
        scale_fn (function): Scale function for learning rates over time. Default is defined by mode.
        scale_mode (str): One of (cycle, iterations). Argument passed to the scale function each step
        gamma (float): Scaling factor for exp_range mode
    """
    def __init__(self, base_lr=0.001, max_lr=0.006, step_size=2000, mode='triangular', scale_fn=None, scale_mode='cycle', gamma=1.):
        super(CyclicLR, self).__init__()

        if not isinstance(base_lr, list):
            base_lr = [base_lr]
        if not isinstance(max_lr, list):
            max_lr = [max_lr]
        if not isinstance(step_size, list):
            step_size = [step_size]

        self.base_lr = base_lr
        self.max_lr = max_lr
        self.step_size = step_size
        self.iterations = 0
        self.optim = None
        self.mode = mode
        self.scale_fn = scale_fn
        self.scale_mode = scale_mode

        if scale_fn is None:
            if mode == 'triangular':
                self.scale_fn = lambda x: 1.0
                self.scale_mode = 'cycle'
            elif self.mode == 'triangular2':
                self.scale_fn = lambda x: 1/(2.**(x-1))
                self.scale_mode = 'cycle'
            elif self.mode == 'exp_range':
                self.scale_fn = lambda x: gamma**x
                self.scale_mode = 'iterations'

    def on_start(self, state):
        super(CyclicLR, self).on_start(state)
        self.optim = state[tb.OPTIMIZER]
        self.iterations = 0

        if len(self.optim.param_groups) > 1:
            if len(self.base_lr) == 1:
                self.base_lr = self.base_lr * len(self.optim.param_groups)
            if len(self.max_lr) == 1:
                self.max_lr = self.max_lr * len(self.optim.param_groups)
            if len(self.step_size) == 1:
                self.step_size = self.step_size * len(self.optim.param_groups)

    def on_sample(self, state):
        super(CyclicLR, self).on_sample(state)
        self.update_lrs()

    def on_step_training(self, state):
        super(CyclicLR, self).on_step_training(state)
        self.iterations = self.iterations + 1

    def update_lrs(self):
        i = 0
        for param_group in self.optim.param_groups:
            param_group['lr'] = self.next_lr(self.iterations, i)
            i += 1

    def next_lr(self, epoch_count, group_id):
        epoch_count = float(epoch_count)
        cycle = np.floor(1.0 + epoch_count / (2.0*self.step_size[group_id]))
        x = np.abs(epoch_count / self.step_size[group_id] - 2.0*cycle + 1.0)
        new_lr = self.base_lr[group_id] + (self.max_lr[group_id]-self.base_lr[group_id])*np.maximum(0.0, (1.0-x))

        scale_param = cycle
        if self.scale_mode != 'cycle':
            scale_param = self.iterations

        return new_lr*self.scale_fn(scale_param)



"""
Main Classes
------------------------------------
..  automodule:: torchbearer.callbacks.imaging.imaging
        :members:
        :undoc-members:

Deep Inside Convolutional Networks
------------------------------------
.. automodule:: torchbearer.callbacks.imaging.inside_cnns
        :members:
        :undoc-members:
"""
from .imaging import *
from .inside_cnns import ClassAppearanceModel, RANDOM



from unittest import TestCase
import torchbearer as tb
from torchbearer.callbacks import lr_finder as lrf
import numpy as np


class TestCyclicLR(TestCase):
    def test_scale_fn(self):

        finder = lrf.CyclicLR(scale_fn=None)
        self.assertTrue(finder.scale_fn(100) == 1.0)

        scaler = lambda x: x**2
        finder2 = lrf.CyclicLR(scale_fn=scaler)
        self.assertTrue(finder2.scale_fn(12) == 144)

    def test_modes(self):
        finder = lrf.CyclicLR(mode='triangular')
        self.assertTrue(finder.scale_mode == 'cycle')
        self.assertTrue(finder.scale_fn(100) == 1.0)

        true_scale_fn = lambda x: 1 / (2. ** (x - 1))
        finder2 = lrf.CyclicLR(mode='triangular2')
        self.assertTrue(finder.scale_mode == 'cycle')
        self.assertTrue(finder2.scale_fn(12) == true_scale_fn(12))

        gamma = 0.9
        true_scale_fn = lambda x: gamma**x
        finder3 = lrf.CyclicLR(mode='exp_range', gamma=gamma)
        self.assertTrue(finder3.scale_mode == 'iterations')
        self.assertTrue(finder3.scale_fn(12) == true_scale_fn(12))

    def test_next_lr_tri(self):
        finder = lrf.CyclicLR(step_size=100)
        self.assertAlmostEqual(finder.next_lr(0, 0), 0.001, places=5)
        self.assertAlmostEqual(finder.next_lr(10, 0), 0.0015000000000000005, places=5)
        self.assertAlmostEqual(finder.next_lr(100, 0), 0.006, places=5)
        self.assertAlmostEqual(finder.next_lr(110, 0), 0.006-0.0005, places=5)
        self.assertAlmostEqual(finder.next_lr(200, 0), 0.001, places=5)

    def test_next_lr_tri2(self):
        scale_fn = lambda x: 1/(2.**(x-1))
        cycle_fn = lambda x: np.floor(1 + x / (2*100))

        finder = lrf.CyclicLR(mode='triangular2', step_size=100)
        self.assertAlmostEqual(finder.next_lr(0, 0), 0.001, places=5)
        self.assertAlmostEqual(finder.next_lr(10, 0), 0.0015000000000000005, places=5)
        self.assertAlmostEqual(finder.next_lr(100, 0), 0.006, places=5)
        self.assertAlmostEqual(finder.next_lr(110, 0), (0.006-0.0005)*scale_fn(cycle_fn(110)), places=5)
        self.assertAlmostEqual(finder.next_lr(200, 0), (0.001)*scale_fn(cycle_fn(200)), places=5)

    def test_next_lr_exp_range(self):
        gamma = 0.9
        scale_fn = lambda x: gamma**x

        finder = lrf.CyclicLR(mode='exp_range', step_size=100, gamma=gamma)
        self.assertAlmostEqual(finder.next_lr(0, 0), 0.001, places=5)
        finder.iterations = 10
        self.assertAlmostEqual(finder.next_lr(10, 0), 0.0015000000000000005*scale_fn(10), places=5)
        finder.iterations = 100
        self.assertAlmostEqual(finder.next_lr(100, 0), 0.006*scale_fn(100), places=5)
        finder.iterations = 110
        self.assertAlmostEqual(finder.next_lr(110, 0), (0.006-0.0005)*scale_fn(110), places=5)
        finder.iterations = 200
        self.assertAlmostEqual(finder.next_lr(200, 0), (0.001)*scale_fn(200), places=5)

    def test_end_to_end(self):
        import torch.optim
        import torch.nn
        model = torch.nn.Linear(10, 10)
        model2 = torch.nn.Linear(10, 10)

        optim = torch.optim.SGD([
                        {'params': model.parameters()},
                        {'params': model2.parameters(), 'lr': 1e-3}
                    ], lr=1e-2, momentum=0.9)

        clr = lrf.CyclicLR(step_size=75, base_lr=[0.001, 0.0001], max_lr=[0.006, 0.0006])
        clr.on_start({tb.OPTIMIZER: optim})

        lrs = []
        for i in range(100):
            clr.on_sample({tb.OPTIMIZER: optim})
            clr.on_step_training({tb.OPTIMIZER: optim})
            for param_group in optim.param_groups:
                lr = param_group['lr']
                lrs.append(lr)

        for i, param_group in enumerate(optim.param_groups):
            lr = param_group['lr']
            self.assertAlmostEqual(lr, 0.004399999999999999*0.1**i)

    def test_end_to_end_2(self):
        import torch.optim
        import torch.nn
        model = torch.nn.Linear(10, 10)
        model2 = torch.nn.Linear(10, 10)

        optim = torch.optim.SGD([
                        {'params': model.parameters()},
                        {'params': model2.parameters(), 'lr': 1e-3}
                    ], lr=1e-2, momentum=0.9)

        clr = lrf.CyclicLR(step_size=[75, 100], base_lr=0.001, max_lr=0.006)
        clr.on_start({tb.OPTIMIZER: optim})

        lrs = []
        for i in range(100):
            clr.on_sample({tb.OPTIMIZER: optim})
            clr.on_step_training({tb.OPTIMIZER: optim})
            for param_group in optim.param_groups:
                lr = param_group['lr']
                lrs.append(lr)

        for i, param_group in enumerate(optim.param_groups):
            lr = param_group['lr']
            if i == 0:
                self.assertAlmostEqual(lr, 0.004399999999999999)
            if i == 1:
                self.assertAlmostEqual(lr, 0.00595)



import torchbearer
from torchbearer import Callback
from torchbearer.callbacks.decorators import once_per_epoch

import torch


def _to_file(filename):
    from PIL import Image

    def handler(image, _):
        ndarr = image.mul(255).clamp(0, 255).byte().permute(1, 2, 0).cpu().numpy()
        im = Image.fromarray(ndarr)
        im.save(filename)
    return handler


def _to_pyplot():
    import matplotlib.pyplot as plt

    def handler(image, _):
        ndarr = image.mul(255).clamp(0, 255).byte().permute(1, 2, 0).cpu().numpy()
        plt.imshow(ndarr)
        plt.show()
    return handler


def _to_tensorboard(name='Image', log_dir='./logs', comment='torchbearer'):
    import torchbearer.callbacks.tensor_board as tb
    import os
    log_dir = os.path.join(log_dir, comment)

    def handler(image, state):
        writer = tb.get_writer(log_dir, _to_tensorboard)
        writer.add_image(name, image.clamp(0, 1), state[torchbearer.EPOCH])
        tb.close_writer(log_dir, _to_tensorboard)
    return handler


def _to_visdom(name='Image', log_dir='./logs', comment='torchbearer', visdom_params=None):
    import torchbearer.callbacks.tensor_board as tb
    import os
    log_dir = os.path.join(log_dir, comment)

    def handler(image, state):
        writer = tb.get_writer(log_dir, _to_visdom, visdom=True, visdom_params=visdom_params)
        writer.add_image(name + str(state[torchbearer.EPOCH]), image.clamp(0, 1), state[torchbearer.EPOCH])
        tb.close_writer(log_dir, _to_visdom)

    return handler


class ImagingCallback(Callback):
    """The :class:`ImagingCallback` provides a generic interface for callbacks which yield images that should be sent to
    a file, tensorboard, visdom etc. without needing bespoke code. This allows the user to easily define custom
    visualisations by only writing the code to produce the image.

    Args:
        transform (callable, optional): A function/transform that  takes in a Tensor and returns a transformed version.
            This will be applied to the image before it is sent to output.
    """
    def __init__(self, transform=None):
        self._handlers = []
        self.transform = (lambda img: img) if transform is None else transform

    def on_batch(self, state):
        raise NotImplementedError

    def process(self, state):
        img = self.on_batch(state)
        if img is not None:
            img = self.transform(img)
            for handler in self._handlers:
                handler(img, state)

    def on_train(self):
        """Process this callback for training batches

        Returns:
            ImagingCallback: self
        """
        _old_step_training = self.on_step_training

        def wrapper(state):
            _old_step_training(state)
            self.process(state)

        self.on_step_training = wrapper
        return self

    def on_val(self):
        """Process this callback for validation batches

        Returns:
            ImagingCallback: self
        """
        _old_step_validation = self.on_step_validation

        def wrapper(state):
            _old_step_validation(state)
            if state[torchbearer.DATA] is torchbearer.VALIDATION_DATA:
                self.process(state)

        self.on_step_validation = wrapper
        return self

    def on_test(self):
        """Process this callback for test batches

        Returns:
            ImagingCallback: self
        """
        _old_step_validation = self.on_step_validation

        def wrapper(state):
            _old_step_validation(state)
            if state[torchbearer.DATA] is torchbearer.TEST_DATA:
                self.process(state)

        self.on_step_validation = wrapper
        return self

    def with_handler(self, handler):
        """Append the given output handler to the list of handlers

        Args:
            handler: A function of image and state which stores the given image in some way

        Returns:
            ImagingCallback: self
        """
        self._handlers.append(handler)
        return self

    def to_file(self, filename):
        """Send images from this callback to the given file

        Args:
            filename (str): the filename to store the image to

        Returns:
            ImagingCallback: self
        """
        return self.with_handler(_to_file(filename))

    def to_pyplot(self):
        """Show images from this callback with pyplot

        Returns:
            ImagingCallback: self
        """
        return self.with_handler(_to_pyplot())

    def to_state(self, key):
        """Put images from this callback in state with the given key

        Args:
            key (StateKey): The state key to use for the image

        Returns:
            ImagingCallback: self
        """
        def handler(img, state):
            state[key] = img
        return self.with_handler(handler)

    def to_tensorboard(self, name='Image', log_dir='./logs', comment='torchbearer'):
        """Direct images from this callback to tensorboard with the given parameters

        Args:
            name (str): The name of the image
            log_dir (str): The tensorboard log path for output
            comment (str): Descriptive comment to append to path

        Returns:
            ImagingCallback: self
        """
        return self.with_handler(_to_tensorboard(name=name, log_dir=log_dir, comment=comment))

    def to_visdom(self, name='Image', log_dir='./logs', comment='torchbearer', visdom_params=None):
        """Direct images from this callback to visdom with the given parameters

        Args:
            name (str): The name of the image
            log_dir (str): The visdom log path for output
            comment (str): Descriptive comment to append to path
            visdom_params (:class:`.VisdomParams`): Visdom parameter settings object, uses default if None

        Returns:
            ImagingCallback: self
        """
        return self.with_handler(_to_visdom(name=name, log_dir=log_dir, comment=comment, visdom_params=visdom_params))


class CachingImagingCallback(ImagingCallback):
    """The :class:`CachingImagingCallback` is an :class:`ImagingCallback` which caches batches of images from the given
    state key up to the required amount before passing this along with state to the implementing class, once per epoch.

    Args:
        key (StateKey): The :class:`.StateKey` containing image data (tensor of size [b, c, w, h])
        transform (callable, optional): A function/transform that  takes in a Tensor and returns a transformed version.
            This will be applied to the image before it is sent to output.
        num_images: The number of images to cache
    """
    def __init__(self,
                 key=torchbearer.INPUT,
                 transform=None,
                 num_images=16):
        super(CachingImagingCallback, self).__init__(transform=transform)
        self.key = key
        self.num_images = num_images

        self._data = None
        self._done = False

    def on_cache(self, cache, state):
        """This method should be implemented by the overriding class to return an image from the cache.

        Args:
            cache (tensor): The collected cache of size (num_images, C, W, H)
            state (dict): The trial state dict

        Returns:
            The processed image
        """
        raise NotImplementedError

    def on_batch(self, state):
        if not self._done:
            data = state[self.key].detach()

            if self._data is None:
                remaining = self.num_images if self.num_images < data.size(0) else data.size(0)

                self._data = data[:remaining]
            else:
                remaining = self.num_images - self._data.size(0)

                if remaining > data.size(0):
                    remaining = data.size(0)

                self._data = torch.cat((self._data, data[:remaining]), dim=0)

            if self._data.size(0) >= self.num_images:
                image = self.on_cache(self._data, state)
                self._done = True
                self._data = None
                return image

    def on_end_epoch(self, state):
        super(CachingImagingCallback, self).on_end_epoch(state)
        self._done = False


class MakeGrid(CachingImagingCallback):
    """The :class:`MakeGrid` callback is a :class:`CachingImagingCallback` which calls make grid on the cache with the
    provided parameters.

    Args:
        key (StateKey): The :class:`.StateKey` containing image data (tensor of size [b, c, w, h])
        transform (callable, optional): A function/transform that  takes in a Tensor and returns a transformed version.
            This will be applied to the image before it is sent to output.
        num_images: The number of images to cache
        nrow: See `torchvision.utils.make_grid <https://pytorch.org/docs/stable/torchvision/utils.html#torchvision.utils.make_grid>`_
        padding: See `torchvision.utils.make_grid <https://pytorch.org/docs/stable/torchvision/utils.html#torchvision.utils.make_grid>`_
        normalize: See `torchvision.utils.make_grid <https://pytorch.org/docs/stable/torchvision/utils.html#torchvision.utils.make_grid>`_
        norm_range: See `torchvision.utils.make_grid <https://pytorch.org/docs/stable/torchvision/utils.html#torchvision.utils.make_grid>`_
        scale_each: See `torchvision.utils.make_grid <https://pytorch.org/docs/stable/torchvision/utils.html#torchvision.utils.make_grid>`_
        pad_value: See `torchvision.utils.make_grid <https://pytorch.org/docs/stable/torchvision/utils.html#torchvision.utils.make_grid>`_
    """
    def __init__(self,
                 key=torchbearer.INPUT,
                 transform=None,
                 num_images=16,
                 nrow=8,
                 padding=2,
                 normalize=False,
                 norm_range=None,
                 scale_each=False,
                 pad_value=0):
        super(MakeGrid, self).__init__(transform=transform)
        self.key = key
        self.num_images = num_images
        self.nrow = nrow
        self.padding = padding
        self.normalize = normalize
        self.norm_range = norm_range
        self.scale_each = scale_each
        self.pad_value = pad_value

    def on_cache(self, cache, state):
        import torchvision.utils as utils
        return utils.make_grid(
            cache,
            nrow=self.nrow,
            padding=self.padding,
            normalize=self.normalize,
            range=self.norm_range,
            scale_each=self.scale_each,
            pad_value=self.pad_value
        )


class FromState(ImagingCallback):
    """The :class:`FromState` callback is an :class:`ImagingCallback` which retrieves and image from state when called.
    The number of times the function is called can be controlled with a provided decorator (once_per_epoch, only_if
    etc.)

    Args:
        key (StateKey): The :class:`.StateKey` containing the image (tensor of size [c, w, h])
        transform (callable, optional): A function/transform that  takes in a Tensor and returns a transformed version.
            This will be applied to the image before it is sent to output.
        decorator: A function which will be used to wrap the callback function. once_per_epoch by default
    """
    def __init__(self, key, transform=None, decorator=once_per_epoch):
        super(FromState, self).__init__(transform=transform)
        self.key = key
        self.on_batch = decorator(self.on_batch)

    def on_batch(self, state):
        try:
            return state[self.key]
        except KeyError:
            return None
